from objectextensions import Extension

from abc import ABC

from ...renderables import Renderable
from ...animations import Animation


class Animated(Extension, ABC):
    @staticmethod
    def can_extend(target_cls):
        return issubclass(target_cls, Renderable)

    @staticmethod
    def extend(target_cls):
        Extension._wrap(target_cls, "__init__", Animated.__wrap_init)

        Extension._set_property(target_cls, "animation", Animated.__animation)
        Extension._set_setter(target_cls, "animation", "animation", Animated.__apply_animation)

        Extension._wrap(target_cls, "update", Animated.__wrap_update)
        Extension._set(target_cls, "generate_animation", Animated.__generate_animation)

    def __wrap_init(self, *args, **kwargs):
        yield
        Extension._set(self, "_animation", self.generate_animation())

    def __animation(self) -> Animation:
        return self._animation

    def __apply_animation(self, animation: Animation):
        """
        Named 'apply' rather than 'set' because it is not guaranteed that the provided animation will be set,
        if it is lower priority than the current animation
        """

        if animation.priority >= self._animation.priority:
            self._animation = animation

    def __wrap_update(self, elapsed_ms, input_events, *args, **kwargs):
        yield
        self._animation.update(elapsed_ms)
        self.surface = self._animation.frame

    def __generate_animation(self) -> Animation:
        """
        Must be overridden.
        Should return a new, correct animation for this renderable based on its current state.
        Note that animations can still be created elsewhere on an ad-hoc basis (e.g. in response to specific events)
        """

        raise NotImplementedError



from pygame import Surface
from managedstate import State
from managedstate.extensions import Registrar

from abc import ABC
from typing import Optional

from ...handlers import HitboxHandler
from ..renderable import Renderable


class Screen(Renderable, ABC):
    """
    Any class that inherits from Screen should be used as a top-level object in the game loop.
    Different subclasses of Screen should represent different views of the game;
    for example, the main menu would be one Screen subclass whereas the game world would be a separate Screen subclass
    """

    def __init__(self, game, state: State.with_extensions(Registrar), surface: Optional[Surface] = None):
        super().__init__(game, surface=surface, render_position=(0, 0), parent=None, priority=None)

        self._state = state
        self._hitbox_handler = HitboxHandler()

        self.register_paths(self._state)

    @property
    def state(self) -> State.with_extensions(Registrar):
        return self._state

    @property
    def hitbox_handler(self) -> HitboxHandler:
        return self._hitbox_handler

    def _update(self, elapsed_ms: int, input_events: list, *args, **kwargs) -> None:
        """
        This method can be further extended as necessary in subclasses
        """

        self._hitbox_handler.reset_checked_collisions()

    @staticmethod
    def register_paths(state: State.with_extensions(Registrar)):
        """
        Can optionally be overridden.
        Register any paths for a state object which would be used by this screen or its children here
        """

        pass



from pygame import Surface

from typing import Any
from datetime import timedelta
from abc import ABC
from weakref import ref


class Animation(ABC):
    def __init__(
            self, parent: "Renderable.with_extensions(Animated)", animation_key: str,
            size: float = 1, speed: float = 1, priority: Any = None
    ):
        self.size: float = size
        self.speed: float = speed

        self._parent_renderable = ref(parent)  # Weakref so that it does not prevent parent object being garbage collected
        self._key = animation_key
        self._priority = priority

        self._elapsed = timedelta(0)
        self._elapsed_effective = timedelta(0)

    @property
    def parent_renderable(self) -> "Renderable.with_extensions(Animated)":
        return self._parent_renderable()

    @property
    def animation_handler(self):
        """
        Shortcut property which accesses the current game screen, and then the current animation handler through that
        """

        return self.parent_renderable.game.animation_handler

    @property
    def key(self) -> str:
        return self._key

    @property
    def priority(self) -> Any:
        return self._priority

    @property
    def elapsed(self) -> timedelta:
        return self._elapsed

    @property
    def elapsed_effective(self) -> timedelta:
        """
        This property tracks elapsed time with animation speed accounted for,
        so that the speed can be changed at any time without causing the animation to lose its correct progress
        """

        return self._elapsed_effective

    @property
    def frame(self) -> Surface:
        """
        Must be overridden with a property that returns a Surface representing
        the current frame of the animation, based on its current state
        """

        raise NotImplementedError

    def update(self, elapsed_ms: int) -> None:
        self._add_elapsed(elapsed_ms)
        self._update(elapsed_ms)

    def _update(self, elapsed_ms: int) -> None:
        """
        Lifecycle method, called automatically each game tick.
        Can optionally be overridden.

        Any additional work that may be needed in specific animations,
        including playing sounds at specific parts of the animation, can be completed here
        """

        pass

    def _add_elapsed(self, elapsed_ms: int) -> None:
        elapsed = timedelta(microseconds=elapsed_ms*1000)
        effective_elapsed = elapsed * self.speed

        self._elapsed += elapsed
        self._elapsed_effective += effective_elapsed



from typing import Tuple
from abc import ABC

from ..stats import Stat, GenericStat
from .renderable import Renderable


class Entity(Renderable, ABC):
    """
    Concrete class tightly coupled to the Room class, which renders a persisting Renderable object in a room.

    This includes people, inanimate objects, decorative objects, objects with no collision etc. and only
    excludes things that would cease to exist once the room is exited
    """

    def __init__(self, parent: "Room", render_position: Tuple[int, int], surface=None, priority=None):
        super().__init__(
            parent.game, parent=parent, surface=surface, render_position=render_position, priority=priority
        )

    @property
    def speed(self) -> Tuple[GenericStat, GenericStat]:
        """
        Should return a pair of stat objects representing x speed and y speed, respectively.

        These stats must be concrete GenericStat objects, to expose a clear write-access interface to any external
        forces
        """

        raise NotImplementedError

    @property
    def acceleration(self) -> Tuple[GenericStat, GenericStat]:
        """
        Should return a pair of stat objects representing x acceleration and y acceleration, respectively.

        These stats must be concrete GenericStat objects, to expose a clear write-access interface to any external
        forces
        """

        raise NotImplementedError

    @property
    def mass(self) -> Stat:
        """
        Used to apply changes in momentum rather than just speed;
        For example, in a collision between two objects there may be a transfer of momentum - the speed
        of one object would be multiplied by its mass, and then divided by the other object's mass before
        being added to that other object's speed
        """

        raise NotImplementedError



from pygame import image


class Methods:
    @staticmethod
    def load_image(file_path: str):
        """
        Wrapper method for pygame.image.load() which takes only a `str` file path, and provides a more informative
        error message when the file cannot be found
        """

        try:
            return image.load(file_path).convert_alpha()
        except FileNotFoundError:
            raise FileNotFoundError(f"unable to locate a file under the following path: {file_path}")



from pygame import event

from typing import Callable, Iterable, Dict, Container


class Config:
    FPS: float = 0  # A value of 0 indicates unlimited framerate
    UPDATES_PER_FRAME: int = 1
    GET_INPUT_EVENTS: Callable[[], Iterable] = event.get

    RESOURCE_FOLDER_PATH = "res"  # Can either be absolute, or relative to the current working directory

    CUSTOM_CLASSES: Dict[str, type] = {}

    HITBOX_TAGS: Container = set()  # Should contain all possible valid tags for the Hitbox class

    ANIMATION_DEFAULT_FPS: float = 50



from typing import FrozenSet, Hashable, Any
from contextlib import contextmanager

from .methods import ErrorMessages
from .stat import Stat


class GenericStat(Stat):
    """
    This class provides a standardised interface for any generic stat, which aims to facilitate the kinds of changes
    and modifiers typically applied to stats via game mechanics.

    It is made up of three main segments - a base value, an added secondary value, and an overall multiplier.
    This can be further broken down:
    - The base and secondary values have two individual multipliers *each*, which are applied to them
      before they are then added together
    - The overall multiplier is also comprised of two individual multipliers, which are both applied
      after the base and secondary values have been added together

    The reason for this structure is that each segment of the formula has one multiplier comprised from
    addition operations, and one multiplier comprised from multiplication operations. This separation in multipliers
    means that order of operations does not matter, so changes to these values may be applied and later reverted freely
    without leaving a permanent effect on the total value.

    Each of these factors is named with clarity for the player in mind - it makes intuitive sense that "base damage"
    and "secondary damage" would be added together when calculating the total, for example.

    If this stat is currently locked (self.is_locked is True or truthy), any attempted changes to these factors
    will raise an exception. Thus, it is important to check this attribute before making any changes,
    and unlock the stat if necessary.

    In practice, each of these attributes should be used as shown below:
    damage = GenericStat()

    # Apply double damage (x2 multiplier)
    damage.overall_multiplied_factor *= 2
    # Remove double damage
    damage.overall_multiplied_factor /= 2

    # Apply 10% increased damage
    damage.overall_summed_factor += 0.1

    # Apply double secondary damage (x2 multiplier)
    damage.secondary_multiplied_factor *= 2

    # Apply 10% increased secondary damage
    damage.secondary_summed_factor += 0.1

    # Apply +2 secondary damage
    damage.secondary_value += 2

    # Apply double base damage (x2 multiplier)
    damage.base_multiplied_factor *= 2

    # Apply 10% increased base damage
    damage.base_summed_factor += 0.1

    # Apply +2 base damage
    damage.base_value += 2
    """

    def __init__(
            self,
            base_value: float = 0, base_summed_factor: float = 1, base_multiplied_factor: float = 1,
            secondary_value: float = 0, secondary_summed_factor: float = 1, secondary_multiplied_factor: float = 1,
            overall_summed_factor: float = 1, overall_multiplied_factor: float = 1,
            is_locked: Any = False
    ):
        self._base_value = base_value
        self._base_summed_factor = base_summed_factor
        self._base_multiplied_factor = base_multiplied_factor

        self._secondary_value = secondary_value
        self._secondary_summed_factor = secondary_summed_factor
        self._secondary_multiplied_factor = secondary_multiplied_factor

        self._overall_summed_factor = overall_summed_factor
        self._overall_multiplied_factor = overall_multiplied_factor

        self._is_locked = is_locked

        self._total = None  # Will be calculated as needed

        self._modified_by = set()
        self._modified_by_frozen = None  # Will be calculated as needed

    def __repr__(self):
        return (
            f"<{type(self).__name__}(base_value={self._base_value}, base_summed_factor={self._base_summed_factor}, "
            f"base_multiplied_factor={self._base_multiplied_factor}, secondary_value={self._secondary_value}, "
            f"secondary_summed_factor={self._secondary_summed_factor}, "
            f"secondary_multiplied_factor={self._secondary_multiplied_factor}, "
            f"overall_summed_factor={self._overall_summed_factor}, "
            f"overall_multiplied_factor={self._overall_multiplied_factor}, is_locked={self._is_locked}) "
            f"modified by {repr(self._modified_by)}>"
        )

    @property
    def base_value(self) -> float:
        return self._base_value

    @base_value.setter
    def base_value(self, value: float):
        """
        This attribute should only have *additions* and *subtractions* applied to it,
        **not** multiplications or divisions
        """

        if self._is_locked:
            ErrorMessages.stat_locked()
        if value == self._base_value:
            return

        self._base_value = value
        self._total = None

    @property
    def base_summed_factor(self) -> float:
        return self._base_summed_factor

    @base_summed_factor.setter
    def base_summed_factor(self, value: float):
        """
        This attribute should only have *additions* and *subtractions* applied to it,
        **not** multiplications or divisions
        """

        if self._is_locked:
            ErrorMessages.stat_locked()
        if value == self._base_summed_factor:
            return

        self._base_summed_factor = value
        self._total = None

    @property
    def base_multiplied_factor(self) -> float:
        return self._base_multiplied_factor

    @base_multiplied_factor.setter
    def base_multiplied_factor(self, value: float):
        """
        This attribute should only have *multiplications* and *divisions* applied to it,
        **not** additions or subtractions
        """

        if self._is_locked:
            ErrorMessages.stat_locked()
        if value == self._base_multiplied_factor:
            return

        self._base_multiplied_factor = value
        self._total = None

    @property
    def secondary_value(self) -> float:
        return self._secondary_value

    @secondary_value.setter
    def secondary_value(self, value: float):
        """
        This attribute should only have *additions* and *subtractions* applied to it,
        **not** multiplications or divisions
        """

        if self._is_locked:
            ErrorMessages.stat_locked()
        if value == self._secondary_value:
            return

        self._secondary_value = value
        self._total = None

    @property
    def secondary_summed_factor(self) -> float:
        return self._secondary_summed_factor

    @secondary_summed_factor.setter
    def secondary_summed_factor(self, value: float):
        """
        This attribute should only have *additions* and *subtractions* applied to it,
        **not** multiplications or divisions
        """

        if self._is_locked:
            ErrorMessages.stat_locked()
        if value == self._secondary_summed_factor:
            return

        self._secondary_summed_factor = value
        self._total = None

    @property
    def secondary_multiplied_factor(self) -> float:
        return self._secondary_multiplied_factor

    @secondary_multiplied_factor.setter
    def secondary_multiplied_factor(self, value: float):
        """
        This attribute should only have *multiplications* and *divisions* applied to it,
        **not** additions or subtractions
        """

        if self._is_locked:
            ErrorMessages.stat_locked()
        if value == self._secondary_multiplied_factor:
            return

        self._secondary_multiplied_factor = value
        self._total = None

    @property
    def overall_summed_factor(self) -> float:
        return self._overall_summed_factor

    @overall_summed_factor.setter
    def overall_summed_factor(self, value: float):
        """
        This attribute should only have *additions* and *subtractions* applied to it,
        **not** multiplications or divisions
        """

        if self._is_locked:
            ErrorMessages.stat_locked()
        if value == self._overall_summed_factor:
            return

        self._overall_summed_factor = value
        self._total = None

    @property
    def overall_multiplied_factor(self) -> float:
        return self._overall_multiplied_factor

    @overall_multiplied_factor.setter
    def overall_multiplied_factor(self, value: float):
        """
        This attribute should only have *multiplications* and *divisions* applied to it,
        **not** additions or subtractions
        """

        if self._is_locked:
            ErrorMessages.stat_locked()
        if value == self._overall_multiplied_factor:
            return

        self._overall_multiplied_factor = value
        self._total = None

    @property
    def is_locked(self) -> Any:
        """
        Determines whether the other attributes of this instance may be modified
        (effectively whether this stat instance is currently mutable or not)

        The value can either be set to True/False to function as a simple *is locked*/*is not locked* flag,
        or can instead be set to a permission level (e.g. using an enum for standardised permissions values)
        which should be matched by any client code that wishes to modify or unlock this stat instance
        """

        return self._is_locked

    @is_locked.setter
    def is_locked(self, value: Any):
        """
        Any source that is setting a value to this property should check first that it supercedes the
        permission level stored in the current value of .is_locked (if its value is not simply True)
        """

        self._is_locked = value

    @property
    def total(self) -> float:
        if self._total is None:
            self._total = (
                (
                    (self._base_value * self._base_summed_factor * self._base_multiplied_factor)
                    + (self._secondary_value * self._secondary_summed_factor * self._secondary_multiplied_factor)
                )
                * self._overall_summed_factor * self._overall_multiplied_factor
            )

        return self._total

    @property
    def modified_by(self) -> FrozenSet[Hashable]:
        if self._modified_by_frozen is None:
            self._modified_by_frozen = frozenset(self._modified_by)

        return self._modified_by_frozen

    def add_modified_by(self, identifier: Hashable) -> None:
        """
        Optional.
        Allows a source to register that it has modified this stat.
        This functionality allows a source to later check if a previous modification it wishes to revert was actually
        applied, or was prevented due to the stat being locked at the time, without needing to store this information
        itself.

        The identifier provided should be unique to the specific instance of modification taking place, as collisions
        with other instances of modification may cause unintended behaviour
        """

        if self._is_locked:
            ErrorMessages.stat_locked()
        if identifier in self._modified_by:
            return

        self._modified_by.add(identifier)
        self._modified_by_frozen = None

    def remove_modified_by(self, identifier: Hashable) -> None:
        """
        Optional.
        Allows a source to register that it has reverted a modification to this stat.

        This is the inverse of .add_modified_by(), and so the same stipulations apply
        """

        if self._is_locked:
            ErrorMessages.stat_locked()
        if identifier not in self._modified_by:
            return

        self._modified_by.remove(identifier)
        self._modified_by_frozen = None

    @contextmanager
    def unlocked(self, permission_level: Any):
        """
        Temporarily unlocks this stat instance to allow modifications to be made, then replaces the lock value as it
        was before unlocking. Requires a value greater than or equal to .is_locked to be given via `permission_level`;
        If this requirement is not met, an exception will be raised
        """

        if permission_level < self._is_locked:
            raise PermissionError(f"insufficient permissions to unlock object: {permission_level} < {self._is_locked}")

        is_locked_value = self._is_locked

        self._is_locked = False
        try:
            yield
        finally:
            self._is_locked = is_locked_value



import pygame
from pygame import Surface

from typing import Optional, Type

from .config import Config
from .constants import Constants
from .handlers import GameEventHandler, GameEventKey, CustomClassHandler, AnimationHandler
from .renderables import Screen


class Game:
    """
    Entry point for your game.
    Requires a pygame display Surface to be passed into the constructor,
    and .screen to be set to a valid Screen object, before calling .start() to begin the game loop
    """

    def __init__(self, window: Surface, config=Config):
        pygame.mixer.pre_init(frequency=Constants.AUDIO_FREQUENCY)
        pygame.init()
        pygame.mixer.init()

        self._window = window
        self._config = config

        self.fps = config.FPS
        self.updates_per_frame = config.UPDATES_PER_FRAME

        self._clock = pygame.time.Clock()
        self._screen = None

        # Game-level handlers
        self._game_event_handler = GameEventHandler()
        self._custom_class_handler = CustomClassHandler(self)
        self._animation_handler = AnimationHandler(self)

    @property
    def window(self) -> Surface:
        return self._window

    @property
    def config(self) -> Type[Config]:
        return self._config

    @property
    def screen(self) -> Optional[Screen]:
        return self._screen

    @screen.setter
    def screen(self, value: Screen):
        if value is self._screen:
            return

        with self._game_event_handler(GameEventKey.CHANGE_SCREEN, self._screen, value):
            self._screen = value

    @property
    def game_event_handler(self) -> GameEventHandler:
        return self._game_event_handler

    @property
    def custom_class_handler(self) -> CustomClassHandler:
        return self._custom_class_handler

    @property
    def animation_handler(self) -> AnimationHandler:
        return self._animation_handler

    def start(self) -> None:
        if self._screen is None:
            raise RuntimeError("a valid Screen object must be set to .screen before the game can be started")

        while True:
            for update_index in range(self.updates_per_frame):
                self._update_screen()

            self._render_screen()

    def _update_screen(self) -> None:
        elapsed_ms = self._clock.tick(self.fps * self.updates_per_frame)
        input_events = list(self.config.GET_INPUT_EVENTS())  # Repackaged into a list to be consumed from it as needed

        with self.game_event_handler(GameEventKey.UPDATE, elapsed_ms, input_events):
            self._screen.update(elapsed_ms, input_events)

    def _render_screen(self) -> None:
        updated_rects = self._screen.render(self._window)

        pygame.display.update(updated_rects)



from typing import Set, FrozenSet, Optional, Callable, Iterable, Union

from ..hitboxes import Hitbox


class HitboxHandler:
    def __init__(self):

        self._hitboxes = set()
        self._hitboxes_by_tag = {}

        self._checked_collisions = set()

    @property
    def checked_collisions(self) -> Set[FrozenSet[Union[Hitbox, "Renderable.with_extensions(Hitboxed)"]]]:
        """
        This set represents every unique hitbox combination that has already been checked for a collision in this tick.
        Each combination is stored here as a frozenset containing the involved hitbox instances
        """

        return self._checked_collisions

    def add(self, hitbox: Hitbox) -> None:
        for tag in hitbox.tags:
            if tag not in self._hitboxes_by_tag:
                self._hitboxes_by_tag[tag] = set()

            self._hitboxes_by_tag[tag].add(hitbox)

        self._hitboxes.add(hitbox)

    def remove(self, hitbox: Hitbox) -> None:
        for tag in hitbox.tags:
            self._hitboxes_by_tag[tag].remove(hitbox)

        self._hitboxes.remove(hitbox)

    def get(
            self,
            tags_any: Optional[Iterable[str]] = None, tags_all: Optional[Iterable[str]] = None,
            custom_filter_key: Optional[Callable] = None
    ) -> FrozenSet[Hitbox]:
        """
        This is a querying method to retrieve a subsection of hitboxes in order to optimise collision checking.
        tags_any defines tags of which at least one must to be present in each returned hitbox, tags_all defines tags
        of which all must be present in each returned hitbox and custom_filter_key will be used to further narrow
        down the results as a more specific filter key.

        If None is provided for each of these parameters, they are simply not used to filter hitboxes down.
        Therefore, providing None for all parameters will result in every hitbox being returned
        """

        if tags_any is None:
            result = set(self._hitboxes)
        else:
            result = set()

            for tag_to_include in tags_any:
                for hitbox_with_tag_to_include in self._hitboxes_by_tag[tag_to_include]:
                    result.add(hitbox_with_tag_to_include)

        if tags_all is None:
            pass
        else:
            hitbox_has_required_tags = lambda hitbox: all(
                required_tag in hitbox.tags for required_tag in tags_all
            )
            result = filter(hitbox_has_required_tags, result)

        if custom_filter_key is not None:
            result = frozenset(filter(custom_filter_key, result))
        else:
            result = frozenset(result)

        return result

    def reset_checked_collisions(self) -> None:
        self._checked_collisions = set()



from typing import Any, Optional
from datetime import timedelta

from .fileanimation import FileAnimation
from ..handlers.enums import AnimationDataKey


class RepeatAnimation(FileAnimation):
    def __init__(
            self, parent: "Renderable.with_extensions(Animated)", animation_key: str,
            size: float = 1, speed: float = 1, priority: Any = None,
            fps: Optional[float] = None, windup_frames: int = 0
    ):
        super().__init__(parent, animation_key, size=size, speed=speed, priority=priority)

        if fps is None:
            fps = self._settings.get(AnimationDataKey.DEFAULT_FPS, None)
        if fps is None:
            fps = self.parent_renderable.game.config.ANIMATION_DEFAULT_FPS

        self._fps = fps
        self._frame_time = timedelta(microseconds=(10**6)/fps)

        # Windup frames are optional non-repeating frames at the start of the animation
        self._windup_frames = windup_frames

    @property
    def fps(self) -> float:
        """
        This property does not factor in animation speed; it is the base framerate of the animation only
        """

        return self._fps

    @property
    def windup_frames(self) -> int:
        return self._windup_frames

    @property
    def frame_index(self):
        frames_elapsed = int(self._elapsed_effective / self._frame_time)

        if frames_elapsed < self._windup_frames:
            return frames_elapsed
        else:
            return (
                (frames_elapsed - self._windup_frames) %
                (self.total_frames - self._windup_frames)
            ) + self._windup_frames



import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"

from .game import Game
from .tagged import Tagged
from .constants import Constants
from .config import Config
from .methods import Methods



from pygame import Surface

from abc import ABC
from typing import Any

from ..handlers.enums import AnimationDataKey
from .animation import Animation


class FileAnimation(Animation, ABC):
    def __init__(
            self, parent: "Renderable.with_extensions(Animated)", animation_key: str,
            size: float = 1, speed: float = 1, priority: Any = None
    ):
        super().__init__(parent, animation_key, size=size, speed=speed, priority=priority)

        self._settings = self.animation_handler.get_settings(type(parent), animation_key)

    @property
    def total_frames(self) -> int:
        return len(self._settings[AnimationDataKey.FRAMES])

    @property
    def frame(self) -> Surface:
        frame_key = self._settings[AnimationDataKey.FRAMES][self.frame_index]

        return self.animation_handler.get_frame(frame_key, self.size)

    @property
    def frame_index(self) -> int:
        """
        Must be overridden with a property that returns the index for
        the current frame of the animation, based on its current state
        """

        raise NotImplementedError



from typing import Callable, Union, Tuple, Literal, Optional
from contextlib import contextmanager


class RemoveListener(Exception):
    pass


class GameEventHandler:
    """
    This class is responsible for notifying any registered listeners when a game event is triggered.
    Game events can be anything related to a change in the game's state (for example - changing the current room,
    a character dying, a spell being cast etc.) and are triggered by calling .on_event()
    with the relevant event_key and any other related information for that event.

    Listeners are categorised by event_key, which is either a string representing the type of event occurring, or
    a tuple containing two strings (the first being either "before" or "after" and the second representing event type).
    A tuple event key such as this will be generated when using this class as a contextmanager;
    in such a case, .on_event() is invoked both before and after the event logic contained in the `with` block, with
    the appropriate tuple being passed as an event key each time.

    Any listener added via only a string for the event type, and not under these tuple variants
    for that same event type, will still be notified when the event is triggered using the above contextmanager.

    Alternatively, event_key can be omitted when adding (or removing) a listener, to indicate that the listener
    should be notified on all events
    """

    def __init__(self):
        self._listeners = {}

    @contextmanager
    def __call__(self, event_type: str, *args, **kwargs):
        self.on_event(("before", event_type), *args, **kwargs)
        yield
        """
        Both the non-specific and the after-specific variants of the event key are passed into `.on_event()` here,
        because any listener stored under the non-specific key should still be called at some point when
        calling listeners stored under the more specific variants (since the relevant event has still taken place)
        """
        self.on_event(event_type, *args, **kwargs)
        self.on_event(("after", event_type), *args, **kwargs)

    def add_listener(
            self, listener: Callable,
            event_key: Optional[Union[str, Tuple[Literal["before", "after"], str]]] = None
    ) -> None:
        self._listeners.setdefault(event_key, set()).add(listener)

    def remove_listener(
            self, listener: Callable,
            event_key: Optional[Union[str, Tuple[Literal["before", "after"], str]]] = None
    ) -> None:
        self._listeners.setdefault(event_key, set()).remove(listener)

    def on_event(self, event_key: Union[str, Tuple[Literal["before", "after"], str]], *args, **kwargs) -> None:
        # Notify listeners registered to all events
        listeners_to_remove = set()
        for listener in self._listeners.get(None, set()):
            try:
                listener(None, *args, **kwargs)
            except RemoveListener:
                listeners_to_remove.add(listener)

        for listener in listeners_to_remove:
            self.remove_listener(listener, None)

        # Notify listeners registered specifically to this event
        listeners_to_remove = set()
        for listener in self._listeners.get(event_key, set()):
            try:
                listener(event_key, *args, **kwargs)
            except RemoveListener:
                listeners_to_remove.add(listener)

        for listener in listeners_to_remove:
            self.remove_listener(listener, event_key)



from sys import modules
from logging import warning


class CustomClassHandler:
    def __init__(self, game):
        self._game = game

        self._classes = {**self._game.config.CUSTOM_CLASSES}

    def register(self, **classes: type) -> None:
        """
        Stores the provided classes under their associated provided names, to be retrieved as needed later.
        This method will output a warning message if stored classes are overwritten, but will not prevent
        them from being overwritten
        """

        for cls_key, cls in classes.items():
            if (cls_key in self._classes) and (self._classes[cls_key] != cls):
                warning(
                    f"{type(self).__name__} has had an entry in its registry overwritten:"
                    f" '{cls_key}' now refers to {cls.__name__} instead of {self._classes[cls_key].__name__}"
                )

        self._classes.update(classes)

    def get(self, class_name: str) -> type:
        """
        Attempts to locate a class using the provided class name.
        First checks locally stored classes (initially populated from Config.CUSTOM_CLASSES
        but additional classes can be manually added during runtime),
        then checks classes in the global namespace (in __main__).
        Supports dot notation to access objects stored under module or class attributes in the global namespace
        """

        if class_name in self._classes:
            return self._classes[class_name]

        try:
            nodes = class_name.split(".")

            result = getattr(modules["__main__"], nodes.pop(0))
            for node in nodes:
                result = getattr(result, node)

            return result

        except AttributeError:
            raise ValueError(f"unable to resolve the class name '{class_name}'")



from pygame import Surface
from managedstate import State
from managedstate.extensions import Registrar
from managedstate.extensions.registrar import PartialQueries

from os import path

from .screen import Screen
from ..room import Room
from ...handlers import GameEventKey
from ...constants import Constants as GameConstants


class World(Screen):
    """
    A concrete implementation of Screen, that represents a standard room-based game world
    """

    def __init__(self, game, state: State.with_extensions(Registrar)):
        super().__init__(game, state)
        self.surface = self.copy_surface()

        # Registering a class `Room` which may appear in the game state and which would be needed inside this class
        self.game.custom_class_handler.register(**{"Room": Room})

        self._current_room = None

        # TODO: Render UI layer after room

    @property
    def curr_room(self) -> Room:
        return self._current_room

    def set_room(self) -> None:
        """
        Checks what room ID the current room should have via the game's state, and if it does not match
        the room ID of the room that is currently active, a new room with the correct room ID is initialised
        and swapped in.

        This method can be manually invoked in order to apply a room change immediately, but this is not strictly
        necessary - when the room ID is changed in the game's state, on the next call to .update() this method will be
        automatically invoked.

        This method assumes that any custom Room subclasses listed in the game's state
        have been made available to the game's CustomClassHandler
        """

        state_current_room_id = self.state.registered_get("current_room_id")
        current_room_id = None if self._current_room is None else self._current_room.room_id

        if state_current_room_id == current_room_id:
            return

        with self.game.game_event_handler(GameEventKey.CHANGE_ROOM, self._current_room, state_current_room_id):
            old_room = self._current_room

            new_room_cls = self.game.custom_class_handler.get(
                self.state.registered_get("room_class", [state_current_room_id])
            )
            new_room = new_room_cls(self, state_current_room_id)

            self._current_room = new_room

            if old_room is not None:
                old_room.parent_recurface = None

    def _update(self, elapsed_ms: int, input_events: list, *args, **kwargs) -> None:
        super()._update(elapsed_ms, input_events, *args, **kwargs)

        self.set_room()

    def copy_surface(self) -> Surface:
        surface = Surface((self.game.window.get_width(), self.game.window.get_height()))
        surface.fill(GameConstants.COLOURS["dev"])

        return surface

    @staticmethod
    def register_paths(state: State.with_extensions(Registrar)):
        state.register_path("current_room_id", ["current_room_id"], [str(None)])

        state.register_path("entity", ["entities", PartialQueries.KEY], [{}])

        state.register_path("room", ["rooms", PartialQueries.KEY], [{}])
        state.register_path("room_class", ["rooms", PartialQueries.KEY, "class"], [{}, {}, "Room"])
        state.register_path("room_entities_ids", ["rooms", PartialQueries.KEY, "entities_ids"], [{}, {}, []])
        state.register_path(
            "room_background_file_path",
            ["rooms", PartialQueries.KEY, "background_file_path"],
            [{}, {}, path.join(f"{Room.__name__}", f"{str(None)}.png")]
        )



from pygame import Surface, transform

from os import path
from json import loads
from typing import Type, Dict, Any, Tuple, Union, Literal

from ..methods import Methods
from .enums import AnimationDataKey


class AnimationHandler:
    """
    Helper class which retrieves and caches files and data for animations
    """

    def __init__(self, game):
        self._game = game

        # The below attributes cache data for performance optimisation
        # Stores animation data which has already been loaded before, by (Animated) class name
        self._animation_data = {}
        # Stores data for any sprite sheets loaded as part of animation data, under the sprite sheet's label
        self._sprite_sheets_data = {}
        # Stores frames which have already been generated, under their frame key & size
        self._frames = {}

    def register_sprite_sheet(
        self,
        sprite_sheet_label: str,
        sprite_data: Dict[
            Literal[AnimationDataKey.FILE_PATH, AnimationDataKey.PARSE_TYPE, AnimationDataKey.PARSE_DATA], Any
        ]
    ) -> None:
        """
        Stores data relating to a sprite sheet, to enable the relevant sprites to be loaded later if necessary
        """

        self._sprite_sheets_data[sprite_sheet_label] = sprite_data

    def get_settings(self, target_cls: Type["Renderable.with_extensions(Animated)"], animation_key: str) -> Dict[str, Any]:
        """
        Retrieves the animation settings associated with the provided class and animation key,
        loading the class's animation data if necessary
        """

        self._load_data(target_cls)

        class_animation_data = self._animation_data[target_cls.__name__]
        class_animation_settings = class_animation_data[AnimationDataKey.ANIMATION_SETTINGS]
        return class_animation_settings[animation_key]

    def get_frame(self, frame_key: Union[str, Tuple[str, int]], size: float = 1) -> Surface:
        """
        Retrieves an animation frame using the provided frame key and size modifier, loading the frame if necessary.

        The frame key should either be a relative path to an image file for the frame
        (relative beginning from the designated resource folder, as indicated in the game's config),
        or it should be a sequence of 2 items
        (the first a sprite sheet label which refers to an already loaded sprite sheet,
        and the second an index for which sprite within in that sprite sheet is the desired frame)
        """

        self._load_frame(frame_key, size)

        return self._frames[frame_key][size]

    def preload_frames(self, target_cls: Type["Renderable.with_extensions(Animated)"], sizes: Tuple[int, ...] = (1,)) -> None:
        """
        Loads animation frames into memory ahead of time.
        Used to prevent bottlenecking in situations where many new frames are needed in a short span of time
        """

        self._load_data(target_cls)

        class_animation_data = self._animation_data[target_cls.__name__]
        class_animation_settings = class_animation_data[AnimationDataKey.ANIMATION_SETTINGS]

        for animation_key, settings in class_animation_settings.items():
            for frame_key in settings[AnimationDataKey.FRAMES]:
                for size in sizes:
                    self._load_frame(frame_key, size)

    def _load_data(self, target_cls: Type["Renderable.with_extensions(Animated)"]) -> None:
        """
        Loads all animation data for the target class, if it is not already loaded.
        Assumes a standard location for the data file as dictated below in the variable `animation_data_file_path`.

        The animation data that is retrieved via this method should include any custom parameters that pertain *only*
        to the animation itself (as opposed to the use case you are implementing it in);
        typically this can include the animation's frames, default framerate etc.

        Sprite sheets are considered unique by label for the purposes of this handler class, meaning that if the same
        sprite sheet label appears in multiple data files, its associated data should be identical.
        To avoid duplication issues resulting from this arrangement, it is recommended to minimise the amount of
        sprite sheets being defined in animation data files, and instead call .register_sprite_sheet() manually
        in your Animated classes' constructors (or elsewhere) as necessary
        """

        if target_cls.__name__ not in self._animation_data:
            animation_data_file_path = path.join(
                self._game.config.RESOURCE_FOLDER_PATH,
                f"{target_cls.__name__}",
                "animation_data.json"
            )

            with open(animation_data_file_path, "r") as file:
                data = loads(file.read())
                sprite_sheets_data = data.get(AnimationDataKey.SPRITE_SHEETS, {})

                self._animation_data[target_cls.__name__] = data
                self._sprite_sheets_data.update(sprite_sheets_data)

    def _load_frame(self, frame_key: Union[str, Tuple[str, int]], size: float) -> None:
        """
        Loads the animation frame at the target size modifier, if it is not already loaded.
        """

        if frame_key not in self._frames or size not in self._frames[frame_key]:
            if type(frame_key) is str:  # frame_key is a file path
                surface = Methods.load_image(
                    path.join(
                        self._game.config.RESOURCE_FOLDER_PATH,
                        frame_key
                    )
                )
                surface = transform.rotozoom(surface, 0, size)

                frame_sizes = self._frames.setdefault(frame_key, {})
                frame_sizes[size] = surface

            else:  # frame_key is a sprite sheet label
                sprite_sheet_label = frame_key[0]
                sprite_index = frame_key[1]

                sprite_sheet_data = self._sprite_sheets_data[sprite_sheet_label]
                pass  # TODO: Add logic to load & store sprites from sprite sheet data at provided zoom



from typing import Type, Tuple, List
from os import path

from ..methods import Methods
from .renderable import Renderable
from .entity import Entity
from .enums import EntityDataKey
from ..extensions import Hitboxed
from ..hitboxes import RecurfaceHitbox


class Room(Renderable.with_extensions(Hitboxed)):
    """
    Concrete class tightly coupled to the World screen, which renders a single room and its occupants.
    Can be subclassed as needed to add further functionality
    """

    def __init__(self, parent: "World", room_id: str):
        super().__init__(parent.game, parent=parent, render_position=(0, 0), priority=0)

        self._room_id = room_id
        self.surface = self._generate_surface()

        self._load_room()

    @property
    def room_id(self) -> str:
        return self._room_id

    def generate_hitboxes(self):
        return [RecurfaceHitbox(self, tags=(), is_inverted=True)]  # TODO: Add relevant tags for Room

    def _generate_surface(self):
        """
        Loads the background image for the room object as a new Surface.
        Assumes a standard location for the image file as dictated below in `background_file_path`
        """

        background_file_path = path.join(
            self.game.config.RESOURCE_FOLDER_PATH,
            self.game.screen.state.registered_get("room_background_file_path", [self._room_id])
        )

        return Methods.load_image(background_file_path)

    def _load_room(self):
        """
        Instantiates all the entities present in the current room.
        Assumes that any custom Entity subclasses listed in the game's state
        have been made available to the game's CustomClassHandler
        """

        curr_room_entities_ids: List[str] = self.game.screen.state.registered_get(
            "room_entities_ids", [self._room_id]
        )

        for entity_id in curr_room_entities_ids:
            entity_data: dict = self.game.screen.state.registered_get("entity", [entity_id])

            entity_class: Type[Entity] = self.game.custom_class_handler.get(entity_data[EntityDataKey.CLASS])
            entity_details: Tuple[list, dict] = (
                entity_data.get(EntityDataKey.ARGS, []),
                entity_data.get(EntityDataKey.KWARGS, {})
            )

            entity_class(self, *entity_details[0], **entity_details[1])



from objectextensions import Decorators

from typing import Iterable, Dict, Type, Callable
from abc import ABC
from weakref import ref

from ..tagged import Tagged


class Hitbox(Tagged, ABC):
    @Decorators.classproperty
    def COLLISION_CHECKERS(cls) -> Dict[Type["Hitbox"], Callable[["Hitbox", "Hitbox"], bool]]:
        """
        When two hitboxes are being checked for a collision, at least one of the two Hitbox objects involved
        should have a compatible collision checker function stored in this property.
        It should be stored using the other involved hitbox's class as a key
        """

        raise NotImplementedError

    def __init__(self, parent: "Renderable.with_extensions(Hitboxed)", tags: Iterable[str] = ()):
        # Weakref so that it does not prevent parent object being garbage collected
        self._parent_renderable = ref(parent)  # Hoisted so that it is available when `._is_valid_tag()` is called

        super().__init__(tags=tags)

    @property
    def hitbox_handler(self) -> "HitboxHandler":
        """
        Shortcut property which accesses the current game screen, and then the current hitbox handler through that
        """

        return self.parent_renderable.game.screen.hitbox_handler

    @property
    def parent_renderable(self) -> "Renderable.with_extensions(Hitboxed)":
        return self._parent_renderable()

    def is_collision(self, other: "Hitbox", check_by_parent: bool = True) -> bool:
        """
        Checks for a collision between this hitbox and the provided other hitbox.
        If this collision has already been checked this tick, returns False.

        If check_by_parent is True, hitbox objects that have the same parent Renderable
        are considered to be the same hitbox for this purpose
        """

        if check_by_parent:
            collision_key = frozenset((self.parent_renderable, other.parent_renderable))
        else:
            collision_key = frozenset((self, other))

        if collision_key in self.hitbox_handler.checked_collisions:
            return False  # This collision check has already been carried out previously this tick

        if type(other) in self.COLLISION_CHECKERS:
            self.hitbox_handler.checked_collisions.add(collision_key)

            return self.COLLISION_CHECKERS[type(other)](self, other)
        elif type(self) in other.COLLISION_CHECKERS:
            self.hitbox_handler.checked_collisions.add(collision_key)

            return other.COLLISION_CHECKERS[type(self)](self, other)
        else:
            raise TypeError(
                "unable to locate a compatible collision checker for a collision between "
                f"`{type(self).__name__}` and `{type(other).__name__}` instances"
            )

    def _is_valid_tag(self, tag: str) -> bool:
        return tag in self.parent_renderable.game.config.HITBOX_TAGS



from abc import ABC
from functools import total_ordering
from operator import add, sub, mul, truediv, floordiv, mod
from typing import Callable, Tuple


@total_ordering
class Stat(ABC):
    @property
    def total(self) -> float:
        raise NotImplementedError

    def __str__(self):
        return f"{type(self).__name__}: {self.total}"

    # Note that @total_ordering covers the comparison methods that are not implemented here
    def __lt__(self, other):
        if issubclass(type(other), Stat):
            return self.total < other.total

        else:
            return self.total < other

    def __eq__(self, other):
        if issubclass(type(other), Stat):
            return self.total == other.total

        else:
            return self.total == other  # Stat objects can be considered comparable to other numerical objects by .total

    def __add__(self, other):
        if not issubclass(type(other), Stat):
            return NotImplemented

        return CombinedStat((self, other), add)

    def __sub__(self, other):
        if not issubclass(type(other), Stat):
            return NotImplemented

        return CombinedStat((self, other), sub)

    def __mul__(self, other):
        if not issubclass(type(other), Stat):
            return NotImplemented

        return CombinedStat((self, other), mul)

    def __truediv__(self, other):
        if not issubclass(type(other), Stat):
            return NotImplemented

        return CombinedStat((self, other), truediv)

    def __floordiv__(self, other):
        if not issubclass(type(other), Stat):
            return NotImplemented

        return CombinedStat((self, other), floordiv)

    def __mod__(self, other):
        if not issubclass(type(other), Stat):
            return NotImplemented

        return CombinedStat((self, other), mod)


class CombinedStat(Stat):
    """
    Represents the result of applying an arithmetic operation on two other stat objects.
    Stores live references to its constituent stat objects so that it may update accordingly when either one changes.

    Each time this stat's total value is calculated, the two constituent stat objects' current values are
    passed (in order) into the provided operator function, and the result is returned.

    These constituent stat objects can also be CombinedStat objects, allowing complex equations to be modeled using
    chains of CombinedStat objects to reach the desired result
    """

    def __init__(self, operands: Tuple[Stat, Stat], operator: Callable[[float, float], float]):
        self._operands = tuple(operands)
        self._operator = operator

        self._operand_totals = (None, None)  # Will be calculated as needed
        self._total = None  # Will be calculated as needed

    @property
    def operands(self) -> Tuple[Stat, Stat]:
        return self._operands

    @property
    def operator(self) -> Callable[[float, float], float]:
        """
        Can be a basic arithmetic operator function, or a custom algorithm, as needed.
        This customisability allows the totals returned from the provided stats to be modified in any way necessary
        within this function, even before combining them together
        """

        return self._operator

    @property
    def total(self) -> float:
        # Pulling fresh operand totals from the referenced stat objects
        operand_totals = (self._operands[0].total, self._operands[1].total)

        # Checking if the total needs recalculating
        # (either the current total is None, or the fresh operand totals don't match the stored ones)
        if (operand_totals == self._operand_totals) and self._total is not None:  # Does not need recalculating
            return self._total

        else:
            self._operand_totals = operand_totals

            self._total = self._operator(*operand_totals)
            return self._total



from enum import Enum


class GameEventKey(str, Enum):
    """
    Constants representing game events which may be triggered
    via calls within this package to GameEventHandler.on_event()
    """

    UPDATE = "update"

    CHANGE_SCREEN = "change_screen"
    CHANGE_ROOM = "change_room"


class AnimationDataKey(str, Enum):
    """
    Constants representing JSON keys which are used in the files containing animation data.

    The structure of data in each file (containing all animation data for one `Animated` class)
    is expected to be as follows:
    {
        "sprite_sheets": {
            "<sprite sheet label>": {
                "file_path": <str value>,
                "parse_type": "individual",
                "parse_data": [
                    [[<int x start>, <int y start>], [<int x end>, <int y end>]],
                    ...
                ]
            },
            ...
        },

        "animation_settings": {
            "<animation key>": {
                "frames": [
                    "<frame key>",
                    "<another frame key>",
                    ...
                ],
                "<animation parameter key>": <value>,
                "<another animation parameter key>": <value>,
                ...
            },

            "<another animation key>": {
                ...
            },
            ...
        }
    }
    """

    # Top-Level Keys
    SPRITE_SHEETS = "sprite_sheets"
    ANIMATION_SETTINGS = "animation_settings"

    # Generic
    FILE_PATH = "file_path"

    # Sprite Sheet Data
    PARSE_TYPE = "parse_type"
    PARSE_TYPE_INDIVIDUAL = "individual"
    PARSE_DATA = "parse_data"

    # Animation Settings
    FRAMES = "frames"  # Should retrieve a list of frame keys
    # Animation-specific default FPS (overrides the generic default animation FPS in the game's config)
    DEFAULT_FPS = "default_fps"  # Should retrieve a float



from objectextensions import Extension

from typing import Iterable, FrozenSet
from abc import ABC

from ...renderables import Renderable
from ...hitboxes import Hitbox


class Hitboxed(Extension, ABC):
    @staticmethod
    def can_extend(target_cls):
        return issubclass(target_cls, Renderable)

    @staticmethod
    def extend(target_cls):
        Extension._wrap(target_cls, "__init__", Hitboxed.__wrap_init)

        Extension._set_property(target_cls, "hitboxes", Hitboxed.__hitboxes)
        Extension._set_setter(target_cls, "hitboxes", "hitboxes", Hitboxed.__set_hitboxes)

        Extension._set(target_cls, "generate_hitboxes", Hitboxed.__generate_hitboxes)
        Extension._set(target_cls, "check_collisions", Hitboxed.__check_collisions)
        Extension._set(target_cls, "collide", Hitboxed.__collide)

    def __wrap_init(self, *args, **kwargs):
        yield
        Extension._set(self, "_hitboxes", frozenset())
        self.hitboxes = self.generate_hitboxes()

    def __hitboxes(self) -> FrozenSet[Hitbox]:
        return self._hitboxes

    def __set_hitboxes(self, hitboxes: Iterable[Hitbox]):
        hitbox_handler = self.game.screen.hitbox_handler

        for hitbox in self._hitboxes:
            if hitbox not in hitboxes:
                hitbox_handler.remove(hitbox)

        for hitbox in hitboxes:
            if hitbox not in self._hitboxes:
                hitbox_handler.add(hitbox)

        self._hitboxes = frozenset(hitboxes)

    def __generate_hitboxes(self) -> Iterable[Hitbox]:
        """
        Must be overridden.
        Should return new, correct hitboxes for this Renderable object based on its current state.
        Note that hitboxes can still be created elsewhere on an ad-hoc basis (e.g. in response to specific events)
        """

        raise NotImplementedError

    def __check_collisions(self) -> None:
        """
        Can optionally be overridden.
        Standardises the entry point for checking collisions for this Renderable object.

        Hitboxes from other Renderable objects that may have valid collisions with this one's hitboxes can be retrieved
        using the get() method on self.game.screen.hitbox_handler, filtered down via the optional parameters
        that method can receive.

        When a collision is detected between two objects, each of their respective collide() methods should
        be invoked and passed the other object involved in that collision.
        """

        pass

    def __collide(self, other: "Renderable.with_extensions(Hitboxed)") -> None:
        """
        Can optionally be overridden.
        Should apply this renderable's on-collision effects to itself and/or the provided other renderable.

        This includes inspecting the other renderable (if needed) to decide what effects should apply.

        When a collision is detected and this method is invoked, it should *also* be invoked on the other object
        involved in the collision and passed this object as an argument
        to allow both objects' collision logic to be applied.

        # TODO: Review the below rule in practice - 'squishy' objects and 'immovable' objects
                    may want to edit the other object, or not be edited, respectively
        - Any *direct* changes to the position or movement (or other stats) of a renderable as a result of the collision
          itself should be handled by that renderable, not by the other involved renderable
        """

        pass



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as readme_file:
    long_description = readme_file.read()

setup(
    name="roomy",
    packages=[
        "roomy",
        "roomy.animations",
        "roomy.extensions", "roomy.extensions.renderable",
        "roomy.handlers", "roomy.hitboxes",
        "roomy.renderables", "roomy.renderables.screens",
        "roomy.stats"
    ],
    version="0.10.0",
    license="MIT",
    description="A game engine specialised for building 2D non-scrolling games, written in python",
    long_description_content_type="text/markdown",
    long_description=long_description,
    author="immijimmi",
    author_email="immijimmi1@gmail.com",
    url="https://github.com/immijimmi/roomy",
    download_url="https://github.com/immijimmi/roomy/archive/refs/tags/v0.10.0.tar.gz",
    keywords=["game", "engine", "2D"],
    install_requires=[
        "pygame~=2.0.1",
        "managedstate~=5.0.0",
        "recurfaces~=2.1.0",
        "objectextensions~=2.0.1"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: pygame",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8"
    ],
)



from .animationhandler import AnimationHandler
from .customclasshandler import CustomClassHandler
from .hitboxhandler import HitboxHandler
from .gameeventhandler import GameEventHandler, RemoveListener
from .enums import GameEventKey, AnimationDataKey



from recurfaces import Recurface
from objectextensions import Extendable
from pygame import Surface

from abc import ABC
from typing import Optional, Tuple, Any


class Renderable(Extendable, Recurface, ABC):
    """
    A Renderable object is any object that has a place in the rendering hierarchy.

    Explanation of additional constructor param requirements:
    - game: A reference to the Game instance allows all entities to navigate the object hierarchy from a static origin
    """

    def __init__(
            self, game: "Game",
            surface: Optional[Surface] = None, render_position: Optional[Tuple[float, float]] = None,
            parent: Optional["Renderable"] = None, priority: Any = None,
    ):
        Extendable.__init__(self)
        Recurface.__init__(self, surface=surface, position=render_position, parent=parent, priority=priority)

        self._game = game

    @property
    def game(self):
        return self._game

    def update(self, elapsed_ms: int, input_events: list, *args, **kwargs) -> None:
        """
        Lifecycle method, called automatically each game tick.

        *args and **kwargs can optionally be passed any additional information regarding the game tick itself;
        for example, a game which wants to prioritise updating the environment first and update the characters second
        in each loop may then choose to run at 2 updates per frame so that the game's screen can first call .update()
        with the kwarg `type="environment"` and second call .update() with the kwarg `type="characters"`.

        Note that additional *args and **kwargs can be introduced at an arbitrary point in the Renderable hierarchy, and
        do not have to be passed in starting from the game's screen object. They also do not have to be propagated
        further down the hierarchy than is necessary
        """

        self._update(elapsed_ms, input_events, *args, **kwargs)

        try:
            # Events passed to high render_priority children first, if possible
            child_recurfaces = reversed(self.ordered_child_recurfaces)
        except TypeError:
            child_recurfaces = self.child_recurfaces

        child: Renderable
        for child in child_recurfaces:
            child.update(elapsed_ms, input_events, *args, **kwargs)

    def _update(self, elapsed_ms: int, input_events: list, *args, **kwargs) -> None:
        """
        Lifecycle method, called automatically each game tick.
        Can optionally be overridden.

        Complete any miscellaneous work necessary for this object each tick
        that is not covered by other lifecycle methods here
        """

        pass



from abc import ABC
from typing import Iterable, FrozenSet


class Tagged(ABC):
    def __init__(self, tags: Iterable[str] = ()):
        for tag in tags:
            if not self._is_valid_tag(tag):
                raise ValueError(f"Invalid tag provided: {tag}")

        self._tags: FrozenSet[str] = frozenset(tags)

    @property
    def tags(self) -> FrozenSet[str]:
        return self._tags

    def _is_valid_tag(self, tag: str) -> bool:
        """
        Must be overridden.
        Should check whether the provided tag is a valid tag for this object's class
        """

        raise NotImplementedError



from objectextensions import Decorators

from typing import Dict, Type, Callable, Iterable

from .hitbox import Hitbox


class RecurfaceHitbox(Hitbox):
    def __init__(self, parent: "Renderable.with_extensions(Hitboxed)", tags: Iterable[str] = (), is_inverted: bool = False):
        super().__init__(parent, tags=tags)

        self._is_inverted = is_inverted

    @property
    def is_inverted(self) -> bool:
        return self._is_inverted

    @Decorators.classproperty
    def COLLISION_CHECKERS(cls) -> Dict[Type[Hitbox], Callable[[Hitbox, Hitbox], bool]]:
        return {
            cls: cls._is_collision_recurfacehitbox
        }

    @staticmethod
    def _is_collision_recurfacehitbox(a: "RecurfaceHitbox", b: "RecurfaceHitbox") -> bool:
        a_left, a_right = (
            a.parent_renderable.x_render_position,
            a.parent_renderable.x_render_position + a.parent_renderable.surface.get_width()
        )
        a_top, a_bottom = (
            a.parent_renderable.y_render_position,
            a.parent_renderable.y_render_position + a.parent_renderable.surface.get_height()
        )

        b_left, b_right = (
            b.parent_renderable.x_render_position,
            b.parent_renderable.x_render_position + b.parent_renderable.surface.get_width()
        )
        b_top, b_bottom = (
            b.parent_renderable.y_render_position,
            b.parent_renderable.y_render_position + b.parent_renderable.surface.get_height()
        )

        if a.is_inverted and b.is_inverted:
            return True
        elif a.is_inverted and (not b.is_inverted):
            return (
                    (b_left < a_left) or (b_right > a_right) or
                    (b_top < a_top) or (b_bottom > a_bottom)
            )
        elif (not a.is_inverted) and b.is_inverted:
            return (
                    (a_left < b_left) or (a_right > b_right) or
                    (a_top < b_top) or (a_bottom > b_bottom)
            )
        else:
            return (
                    (a_left < b_right) and (b_left < a_right) and
                    (a_top < b_bottom) and (b_top < a_bottom)
            )
